#preprocess_data.py

import xarray as xr
import xesmf as xe
import numpy as np


def rename_dims_to_standard(ds):
    """
    Automatically detect and rename dimensions to the standard names: 'time', 'lat', 'lon'.
    
    ds (xarray.Dataset or xarray.DataArray): Input dataset or data array.
    
    """
    
    rename_dict = {}

    # Step 1: Detect longitude
    if 'lon' not in ds.dims:
        for possible_lon in ['longitude', 'x', 'lon', 'LONGITUDE', 'Longitude']:
            if possible_lon in ds.dims or possible_lon in ds.coords:
                rename_dict[possible_lon] = 'lon'
                break

    # Step 2: Detect latitude
    if 'lat' not in ds.dims:
        for possible_lat in ['latitude', 'y', 'lat', 'LATITUDE', 'Latitude']:
            if possible_lat in ds.dims or possible_lat in ds.coords:
                rename_dict[possible_lat] = 'lat'
                break

    # Step 3: Detect time
    if 'time' not in ds.dims:
        for possible_time in ['time', 't', 'TIME', 'Time']:
            if possible_time in ds.dims or possible_time in ds.coords:
                rename_dict[possible_time] = 'time'
                break

    # Step 4: Rename the detected dimensions
    if rename_dict:
        ds = ds.rename(rename_dict)
    
    return ds



def adjust_longitude(ds, lon_name='lon', to_range='0_360'):
    
    ''' Adjust the longitude values either from [-180, 180] to [0, 360] or from [0, 360] to [-180, 180].
    
    ds : Input dataset or data array.
    lon_name : Name of the longitude variable (default is 'lon').
    to_range : The target longitude range, either '0_360' or '-180_180'. '''
    
    lon = ds[lon_name]
    
    if to_range == '0_360':

        if lon.min() < 0:
            ds[lon_name] = lon.where(lon >= 0, lon + 360)
    elif to_range == '-180_180':

        if lon.max() > 180:
            ds[lon_name] = lon.where(lon <= 180, lon - 360)
    else:
        raise ValueError("Invalid target range. Use '0_360' or '-180_180'.")

    ds = ds.sortby(lon_name)
    
    return ds





# def load_data(path, var=None, start_time=None, end_time=None, lat_s=None, lat_e=None, lon_s=None, lon_e=None):

# 	''' This function loads the variables from nc files as xarray.DataArray 
# 		path: directory to nc file
# 		var : variables to be extracted
# 		start_time: start time for the variables to be extracted
# 		end_time: end time for the variables to be extracted   
# 		(end_time - start_time = time period for analysis)
# 		lat_s: start latitude
# 		lat_e: end latitude (check latitude order using ncdump -v lat ncfile.nc 
# 		lat_s and lat_e should be choosen in earlier and later latidues, in the 
# 		output of ncdump, both should be provided or none for global)
# 		lon_s: start longitude
# 		lon_e: end longitude (check longitude order using ncdump -v lon ncfile.nc 
# 		lon_s and lon_e should be choosen in order of earlier and later longidues, 
# 		in the output of ncdump, both should be provided or none for global)'''


#     if var == None:
#         return xr.open_dataset(path)

#     else:
#         if (start_time == None or end_time==None) and (lat_s==None or lat_e==None) and(lon_s==None or lon_e==None):
#             return xr.open_dataset(path)[var]
#         elif (lat_s==None or lat_e==None) and (lon_s==None or lon_e==None):
#             return xr.open_dataset(path)[var].sel(time=slice(f'{start_time}-01-01', f'{end_time}-12-30'))

#         elif (start_time == None or end_time==None) and(lon_s==None or lon_e==None):
#             return xr.open_dataset(path)[var].sel(lat=slice(lat_s, lat_e))

#         elif (start_time == None or end_time==None) and(lat_s==None or lat_e==None):
#             return xr.open_dataset(path)[var].sel(lon=slice(lon_s, lon_e))

#         elif (start_time == None or end_time==None):
#             return xr.open_dataset(path)[var].sel(lon=slice(lon_s, lon_e), lat=slice(lat_s, lat_e))
#         elif (lat_s==None or lat_e==None):
#             return xr.open_dataset(path)[var].sel(lon=slice(lon_s, lon_e), time=slice(f'{start_time}-01-01', f'{end_time}-12-30'))

#         elif (lon_s==None or lon_e==None):
#             return xr.open_dataset(path)[var].sel(lat=slice(lat_s, lat_e), time=slice(f'{start_time}-01-01', f'{end_time}-12-30'))

#         else:
#             return xr.open_dataset(path)[var].sel(lat=slice(lat_s, lat_e), time=slice(f'{start_time}-01-01', f'{end_time}-12-30'), lon=slice(lon_s, lon_e))


def load_data(path, var=None, start_time=None, end_time=None, lat_s=None, lat_e=None, lon_s=None, lon_e=None):
    """
    Load variables from NetCDF files as xarray.DataArray or Dataset.
    
    Parameters:
    -----------
    path : str
        Path to the NetCDF file.
    var : str, optional
        Variable name to extract from the file. If None, the entire dataset is returned.
    start_time : str, optional
        Start year in 'YYYY' format for the time range selection.
    end_time : str, optional
        End year in 'YYYY' format for the time range selection.
    lat_s : float, optional
        Start latitude for spatial selection. If None, the full latitude range is selected.
    lat_e : float, optional
        End latitude for spatial selection. If None, the full latitude range is selected.
    lon_s : float, optional
        Start longitude for spatial selection. If None, the full longitude range is selected.
    lon_e : float, optional
        End longitude for spatial selection. If None, the full longitude range is selected.

    Returns:
    --------
    xarray.DataArray or xarray.Dataset
        The selected subset of data based on the provided parameters.
    """
    

    ds = xr.open_dataset(path)
    

    if var:
        ds = ds[var]
    

    selection = {}


    if start_time and end_time:
        selection['time'] = slice(f'{start_time}-01-01', f'{end_time}-12-31')


    if lat_s is not None and lat_e is not None:
        selection['lat'] = slice(lat_s, lat_e)
    

    if lon_s is not None and lon_e is not None:
        selection['lon'] = slice(lon_s, lon_e)


    if selection:
        ds = ds.sel(**selection)
    
    return ds



def write_netcdf(data, file_path):
    """Save the given data to a NetCDF file.
    data: xarray.DataArray to be saved on netcdf format
    file_path: directory where file needs to be saved"""
    data.to_netcdf(file_path)
    print(f"Data saved to {file_path}")


import xarray as xr
import xesmf as xe

def regridding(ds, ds_out=None, var=None, method=None, to_range=None, x_s=None, x_e=None, x_i=None, y_s=None, y_e=None, y_i=None):
    '''
    This function is useful for regridding rectilinear and curvilinear grids.
    
    ds : xarray.Dataset or xarray.DataArray
        The dataset or dataarray containing the variable grid to be regridded.
        
    ds_out : xarray.Dataset or xarray.DataArray
        The target dataset or dataarray containing the target grid.
        if ds_out==None, regridding will be done using a generic grid made using xe, xs, ye, ys, xi and yi

        
    var :
        The name of the variable to be regridded. If None, it is assumed that `ds` and `ds_out` are DataArrays.
        
    to_range :
        The target longitude range, either '0_360' or '-180_180'. Default is '0_360'.
        
    method : str, optional
        The method for regridding. Default is 'bilinear'. Other values can be:
        ['bilinear', 'conservative', 'patch', 'conservative_normed', 'nearest_s2d', 'nearest_d2s'].
        For more details, see:
        - https://xesmf.readthedocs.io/en/stable/notebooks/Compare_algorithms.html
        - https://earthsystemmodeling.org/esmpy_doc/release/latest/ESMPy.pdf

    x_s: start longitude
    x_e: end longitude'
    x_i: lon incriment

    y_s: start latitude
    y_e: end latitude'
    y_i: lat incriment

    To perform regridding from Curvilinear data to rectilinear data we suggest to provide 
    Curvilinear as ds and rectilinear as ds_out
    '''
    if ds_out==None:

        

        if x_s==None and x_e==None and y_s==None and y_e==None:
            x_s=0
            x_e=360
            y_s=90
            y_e=-90
        if x_i==None:
            x_i=1
        if y_i==None:
            y_i=-1

        if isinstance(ds, xr.Dataset):
            ds_out = xr.Dataset(
            {
                "lat": (["lat"], np.arange(y_s, y_e+y_i, y_i)),
                "lon": (["lon"], np.arange(x_s, x_e+x_i, x_i)),
            }
            )

        elif isinstance(ds, xr.DataArray):
            ds_out = xr.DataArray(
            coords=[np.arange(y_s, y_e+y_i, y_i), np.arange(x_s, x_e+x_i, x_i)],
            dims=["lat", "lon"],
            )

        else:
            raise TypeError("Both ds should be either xarray.DataArray or xarray.Dataset.")

    if to_range is None:
        to_range = '0_360'
    if method is None:
        method = 'bilinear'
    
    regridder = xe.Regridder(ds, ds_out, method=method)


    if isinstance(ds, xr.DataArray) and isinstance(ds_out, xr.DataArray):

        regridded = regridder(ds)
    elif isinstance(ds, xr.Dataset) and isinstance(ds_out, xr.Dataset):

        # if var is None:
        #     raise ValueError("Please specify 'var' when regridding a Dataset.")
        regridded = regridder(ds)#[var])
    else:
        raise TypeError("Both ds and ds_out should be either xarray.DataArray or xarray.Dataset.")
    

    return adjust_longitude(rename_dims_to_standard(regridded), to_range=to_range, lon_name='lon')







# anomaly.py

import xarray as xr
import numpy as np
from eofs.xarray import Eof
import xeofs
import matplotlib.pyplot as plt
import cartopy
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from cartopy.util import add_cyclic_point#(data, coord=None, axis=-1) 


def compute_weights(data, lat_dim=None):
    if lat_dim==None:
        lat_dim='lat'
    """Compute cosine weights based on latitude."""
    return np.cos(np.deg2rad(data[f'{lat_dim}']))

def calculate_anomaly(data, clim_start=None, climatology_dim='time', clim_end=None, freq='month'):   
    """Calculate anomaly based on climatology."""
    if clim_start is None:
        if freq == 'month':
            return data.groupby(f'{climatology_dim}.{freq}') - data.groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        elif freq == 'dayofyear':
            return data.groupby(f'{climatology_dim}.{freq}') - data.groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        else:
            raise ValueError('Frequency must be "month" or "dayofyear".')
    else:
        climatology = data.sel(time=slice(f'{clim_start}-01-01', f'{clim_end}-12-31')).groupby(f'{climatology_dim}.{freq}').mean((f'{climatology_dim}'))
        return data.groupby(f'{climatology_dim}.{freq}') - climatology


def compute_eofs(data, lat_dim='lat'):

    ''' This functions computes EOF using eofs module '''

    coslat = np.cos(np.deg2rad(data.coords[lat_dim].values))
    weights = np.sqrt(coslat)

    if weights.ndim == 1 and len(weights) == data.sizes[lat_dim]:
        # Expand weights for broadcasting
        weights = weights[..., np.newaxis]  # Add a new axis for broadcasting in EOF calculation
    else:
        raise ValueError("Weights dimensions do not align with the data anomaly.")

    try:
        if lat_dim not in data.dims:
            solver = Eof(data)
        else:
            solver = Eof(data, weights=weights)
        return solver
        
    except Exception as e:
        print(f"Error creating EOF solver: {e}")
        return None  # Return None on failure to avoid confusion with the exception message

def line_plot(data, figsize=None, dpi=None, variance_fraction=None, color=None, label=None):

    ''' This function plots the calculated indices for the priliminary check 
        if the user wants.
        data: 1D data to be plot
        figsize: size of the canvas. Tuple e.g. (10, 3) default
        dpi : dots per inches-square. Integer 300 default
        variance_fraction: Variance fraction calculated from the given functions 
        such as in calculate_pdo etc.
        color: color for the line desired, black if None
        label: label for the line to be plotted, e.g. PDO, ENSO etc.'''
    try:
        if data.ndim == 1 and 'time' in data.dims:
            plt.figure(figsize=figsize, dpi=dpi)
            ax = plt.axes()
            if ax==None:
                plt.figure(figsize=(10, 3), dpi=300)
                ax=plt.axes()
            if color==None:
                color='k'   
            data.plot(ax=ax, color=color, lw=1, label=label)
            if variance_fraction is not None:
                y_position = data.max() - (data.max() * 0.05) 
                ax.text(int(str(data.time.data[0])[:4]), y_position, f'{label}_vf: {"%.2f" %variance_fraction}')
            # ax.set_title(f'{label}', loc=title_loc, color=color)
            ax.set_title('', loc='center', color='w')
            ax.legend(frameon=False)
            plt.show()
    except Exception as e:
        return print(f"Error while plotting: {e}")



def contour_plot(data, projection=None, figsize=None, dpi=None, cmap=None, extend=None, levels=None, central_lon=None, central_lat=None):

    ''' This function plots the calculated variability patterbs for the 
        priliminary check if the user wants.
        data (xr_DataArray): 2D data to be plot
        projection (projection object ccrs.Projection_type()): cartopy projection for contour plot
        cmap (str): colormap
        extend (str): extend colorbar, takes arguments, max, min, both 
        figsize (tuple; e.g. default (12, 6)): size of the canvas.
        dpi (int default 450): dots per inches-square
        central_lat: center of latitude for projection
        central_lon: centre of longitude for projection
        levels (list sorted as min to max.)= contour levels

        '''

    try:
        if data.ndim==2 and 'lat' in data.dims and 'lon' in data.dims:
            if central_lon==None:
                central_lon=0
            if central_lat==None:
                central_lat=0
            if projection==None:
                projection=ccrs.PlateCarree(central_longitude=central_lon)
            plt.figure(figsize=figsize, dpi=dpi)
            ax = plt.axes(projection=projection)
            lon=data.lon
            if levels == None:
                levels=list(np.linspace(data.min(), data.max(), 10))
            if (lon.min() <= -178 and lon.max() >= 178) or (lon.min() <= 2. and lon.max() >= 357.):
                cyclic_data, cyclic_lon = add_cyclic_point(data, coord=lon)

                cyclic_data = xr.DataArray(
                    cyclic_data, 
                    dims=data.dims,  # Keep the same dimensions
                    coords={**data.coords, 'lon': cyclic_lon},  # Replace the original longitude with cyclic longitude
                    attrs=data.attrs  # Copy the original attributes
                )
                ax.set_global()
                 
                data = cyclic_data#.plot.contourf(levels=levels, cmap=cmap, extend=extend, transform=ccrs.PlateCarree())
            # else:
            data.plot.contourf(levels=levels, cmap=cmap, extend=extend, transform=ccrs.PlateCarree())
            # ax.coastlines()
            # ax.cfeature()
            ax.add_feature(cfeature.LAND)
            ax.add_feature(cfeature.OCEAN)
            ax.add_feature(cfeature.COASTLINE)
    except Exception as e:
        return print(f"Error while plotting: {e}")



def compute_rotated_eofs(data, rotated=None, n_modes=10, standardize=None, use_coslat=None):
    ''' This functions computes EOF using eofs module 

        data: xarray.DataArray with time dimension
        rotated (str): Unrotated EOFs if None, otherwise 
        varimax for orthogonal rotations or Promax for Oblique
        use_coslat (boolean): True if None, otherwise give False. It is used to calculated area-averaged EOFs
        standardize (Boolean) = False if None, make it True if you want to standardize
        n_modes (int >0): change accordingly''' 

    if standardize==None:
        standardize=False
    if use_coslat==None:
        use_coslat=True

    if rotated == None:
        rotated=False
    model = xeofs.single.EOF(n_modes=n_modes, standardize=standardize, use_coslat=use_coslat)
    model.fit(data, dim="time")
    try:
        if rotated==False:
            return model

        if rotated == 'Varimax':
            rot_model = xeofs.single.EOFRotator(n_modes=n_modes, power=1)  ## power=1, 'Varimax (Orthogonal) rotation'
            return rot_model.fit(model)
        elif rotated == 'Promax':
            rot_model = xeofs.single.EOFRotator(n_modes=n_modes, power=4)  ## power=1, 'Promax (Obligue) rotation'
            return rot_model.fit(model)
        else:
            print('Incorrect statement for rotated, give only None or Varimax or Promax')        
        return None



    except Exception as e:
        print(f"Error creating EOF solver: {e}")
        return None  # Return None on failure to avoid confusion with the exception message



# eof_calculations.py

import numpy as np
from .utils import calculate_anomaly, compute_weights, compute_eofs, compute_rotated_eofs
import xarray as xr
from .preprocess_data import load_data, rename_dims_to_standard, adjust_longitude



def calculate_global_mean_sst(data, lat_name='lat', lon_name='lon'):
    return calculate_anomaly(data).weighted(compute_weights(data)).mean(dim=[lat_name, lon_name])



def global_sst_trend_and_enso(data=None, path=None, var=None ,clim_start=None, clim_end=None, desired=None, start_time=None, end_time=None, to_range=None, standardize=None, normalize_pattern=None, normalize_index=None):
    '''This function calculates global sst warming trend, global sst warming pattern
        ENSO index, ENSO pattern and their corrosponding variance fraction.
        They are calculated as First (warming pattern and trend) and Second (ENSO 
        pattern and ENSO index) EOF modes of SST anomaly after removing
        the Annualcycle climatology.

        data: Processed gridded SST data in the shape of (time, lat, lon)
        if not in the order, can be reordered as xarray.transpose
        path (directory): path to data
        var (str): variable name
        clim_start: climatology start year
        clim_end: climatology end year (if both None, whole period will be chosen to compute climatology)
        desired: desired variables, can be choosen from 
        ['sst_trend_pattern', 'sst_trend_timeseries', 'variance_fraction_trend', 
        'enso_pattern', 'enso_index', 'variance_fraction_enso']
        start_time (int): start year for the data to be visulaized
        end_time (int): end year for the data to be visulaized
        to_range (str default 0_360): Range of longitude for the data, 0_360 or -180_180
        standardize (boolean): Normalize the data with Standard Deviation
        Normalize_pattern (boolean): Normalize components (spatial Pattern) with singular values
        Normalize_index (boolean): Normalize scores (timeseries) with singular values'''    
    
    if desired is None:
        desired = ['sst_trend_pattern', 'sst_trend_timeseries', 'variance_fraction_trend','enso_pattern', 'enso_index', 'variance_fraction_enso']
    if standardize==None:
        standardize=False
    if normalize_index==None:
        normalize_index = False
    if normalize_pattern==None:
        normalize_pattern = True
    
    if to_range==None:
        to_range='0_360'    
    if path is not None and var is not None:
        if to_range==None:
            to_range='0_360' ## otherwise provide -180_180
        data = adjust_longitude(rename_dims_to_standard(load_data(path, var, start_time=start_time, end_time=end_time)), to_range=to_range)
        data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    elif data is not None:
        data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    solver = compute_rotated_eofs(data, rotated=False, n_modes=2, standardize=None, use_coslat=True)
    # solver=compute_eofs(data_anom, lat_dim='lat')
    return_desired = []
    # eofs_ = solver.eofs(neofs=2, eofscaling=2)
    # pcs_  = solver.pcs(npcs=2, pcscaling=1)
    eofs_ = solver.components(normalized=normalize_pattern)
    pcs_ = solver.scores(normalized=normalize_index)
    var_frac_ = solver.explained_variance_ratio()
    # var_frac_ = solver.varianceFraction(neigs=2)
    if 'sst_trend_pattern' in desired:
        return_desired.append(eofs_[0].squeeze())
    if 'sst_trend_timeseries' in desired:
        return_desired.append(pcs_[0].squeeze())
    if 'variance_fraction_trend' in desired:
        return_desired.append(var_frac_[0].squeeze())
    if 'enso_pattern' in desired:
        return_desired.append(eofs_[1].squeeze())
    if 'enso_index' in desired:
        return_desired.append(pcs_[1].squeeze())
    if 'variance_fraction_enso' in desired:
        return_desired.append(var_frac_[1].squeeze())

    
    if len(return_desired)==1:
        return return_desired[0]
    else:
        return return_desired



def compute_regional_eof_modes(data=None, path=None, var=None ,clim_start=None, clim_end=None, desired=None, start_time=None, end_time=None, 
    lat_s=None, lat_e=None, lon_s=None, lon_e=None, to_range=None, n_modes=None, remove_trend=None, rotated=None, use_coslat=None, standardize=None, 
    normalize_pattern=None, normalize_index=None):
    '''This function is more generalized form of global_sst_trend_and_enso function.
        
        data: Processed gridded SST data in the shape of (time, lat, lon)
        if not in the order, can be reordered as xarray.transpose
        path (directory): path to data
        var (str): variable name
        desired: desired variables, can be choosen from 
        ['regional_patterns', 'regional_timeseries', 'variance_fractions_regional']
        start_time (int): start year for the data to be visulaized
        end_time (int): end year for the data to be visulaized
        to_range (str default 0_360): Range of longitude for the data, 0_360 or -180_180
        lat_s = start latitude
        lat_e = end latitude
        lon_s = start longitude
        lon_e = end longitude
        standardize (boolean): Normalize the data with Standard Deviation
        Normalize_pattern (boolean): Normalize components (spatial Pattern) with singular values
        Normalize_index (boolean): Normalize scores (timeseries) with singular values
        n_modes (int): number of modes to be calculated
        remove_trend (boolean): Remove global mean variable (a timeseries)
        rotated (str): Unrotated EOFs if None, otherwise 
        Varimax for orthogonal rotations or Promax for Oblique
        use_coslat (boolean): True if None, otherwise give False. It is used to calculated area-averaged EOFs
        clim_start (int): Climatology start year
        clim_end (int): Climatology end year  (if both None, whole period will be chosen to compute climatology)
'''

    if lat_s==None and lat_e==None and lon_s==None and lon_e==None:
        lat_s = 90
        lat_e = -90
        lon_s = 0
        lon_e = 360

    if rotated ==None:
        rotated = False

    if standardize==None:
        standardize=False

    if use_coslat==None:
        use_coslat=True

    if normalize_index==None:
        normalize_index = False
    if normalize_pattern==None:
        normalize_pattern = True
    
    if remove_trend==None:
        remove_trend==False    
    
    if desired is None:
  
        desired = ['regional_patterns', 'regional_timeseries', 'variance_fractions_regional']

    if to_range==None:
        to_range='0_360'    
    if path is not None and var is not None:
        if to_range==None:
            to_range='0_360' ## otherwise provide -180_180
        data = adjust_longitude(rename_dims_to_standard(load_data(path, var, start_time=start_time, end_time=end_time, lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e)), to_range=to_range)
        # data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    elif data is not None:
        data = data
    if remove_trend==False:

        data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    else:
        data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end) - calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
    solver = compute_rotated_eofs(data, rotated=rotated, n_modes=n_modes, standardize=standardize, use_coslat=use_coslat)
    # solver=compute_eofs(data_anom, lat_dim='lat')
    return_desired = []
    if 'regional_patterns' in desired:
        return_desired.append(solver.components(normalized=normalize_pattern).squeeze())
    if 'regional_timeseries' in desired:
        return_desired.append(solver.scores(normalized=normalize_index).squeeze())
    if 'variance_fractions_regional' in desired:
        return_desired.append(solver.explained_variance_ratio().squeeze())

    if len(return_desired)==1:
        return return_desired[0]
    else:
        return return_desired




def compute_pdo(data=None, path=None, var=None ,clim_start=None, clim_end=None, desired=None, start_time=None, end_time=None, to_range=None, 
    standardize=None, normalize_pattern=None, normalize_index=None, lat_s=None, lat_e=None, lon_s=None, lon_e=None, remove_trend=None):

    ''' This function calculates conventional PDO index and PDO pattern and 
        variance fraction.
        It is calculated as First EOF modes of SST anomaly in North Pacific above 20N 
        after removing the Annualcycle climatology and global mean SST (warming trend).

        data: Processed gridded SST data in the shape of (time, lat, lon)
        if not in the order, can be reordered as xarray.transpose
        path (directory): path to data
        var (str): variable name
        clim_start: climatology start year
        clim_end: climatology end year (if both None, whole period will be chosen to compute climatology)
        desired: desired variables, can be choosen from 
        ['pdo_pattern', 'pdo_index', 'variance_fraction_pdo'] 
        start_time (int): start year for the data to be visulaized
        end_time (int): end year for the data to be visulaized
        to_range (str default 0_360): Range of longitude for the data, 0_360 or -180_180
        standardize (boolean): Normalize the data with Standard Deviation
        Normalize_pattern (boolean): Normalize components (spatial Pattern) with singular values
        Normalize_index (boolean): Normalize scores (timeseries) with singular values
        remove_trend (boolean): default is False and PDO index is mode 2, otherwise mode 1.

        lat_s = start latitude
        lat_e = end latitude
        lon_s = start longitude
        lon_e = end longitude
        latitude and longitude ranges can be choosen to experience various versions of PDO'''
    n_modes=None

    if lat_s==None and lat_e==None and lon_s==None and lon_e==None:
        lat_s = 70
        lat_e = 20
        lon_s = 110
        lon_e = 260
    if remove_trend==None or remove_trend==False:
        remove_trend==False
        n_modes=2
    elif remove_trend==True:
        n_modes=1
    else:
        print('remove_trends takes only boolean arguments, i.e. True or False')
    if standardize==None:
        standardize=False
    if normalize_index==None:
        normalize_index = False
    if normalize_pattern==None:
        normalize_pattern = True
    if to_range==None:
        to_range='0_360'    ## otherwise provide -180_180


    if desired is None:
        desired = ['pdo_pattern', 'pdo_index', 'variance_fraction_pdo']

    if path is not None and var is not None:

        data = adjust_longitude(rename_dims_to_standard(load_data(path, var, start_time=start_time, end_time=end_time, lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e)), to_range=to_range)
        # data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    elif data is not None:
        data = data


    data_anomaly = calculate_anomaly(data)
    data_pdo = data.sel(lat=slice(lat_s, lat_e), lon=slice(lon_s, lon_e))
    if remove_trend==remove_trend:
        data_pdo_anomaly = calculate_anomaly(data_pdo, clim_start=clim_start, clim_end=clim_end)
    else:
        data_pdo_anomaly = calculate_anomaly(data_pdo, clim_start=clim_start, clim_end=clim_end) - calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
    # solver=compute_eofs(data_pdo_anomaly, lat_dim='lat')
    solver = compute_rotated_eofs(data_pdo_anomaly, rotated=False, n_modes=1, standardize=None, use_coslat=True)
    return_desired = []
    if remove_trend==False:
        if 'pdo_pattern' in desired:
            return_desired.append(solver.components()[n_modes-1].squeeze())
        if 'pdo_index' in desired:
            return_desired.append(solver.scores()[n_modes-1].squeeze())
        if 'variance_fraction_pdo' in desired:
            return_desired.append(solver.explained_variance_ratio()[n_modes-1].squeeze())
    else:
        if 'pdo_pattern' in desired:
            return_desired.append(solver.components().squeeze())
        if 'pdo_index' in desired:
            return_desired.append(solver.scores().squeeze())
        if 'variance_fraction_pdo' in desired:
            return_desired.append(solver.explained_variance_ratio().squeeze())

    if len(return_desired)==1:
        return return_desired[0]
    else:
        return return_desired




def compute_amo(path=None, var=None, data=None, clim_start=None, clim_end=None, desired=None, start_time=None, end_time=None, lat_s=None, lat_e=None, lon_s=None, lon_e=None, to_range=None):

    if lat_s==None and lat_e==None and lon_s==None and lon_e==None:
        lat_s = 70
        lat_e = 0
        lon_s = 280
        lon_e = 360
    ''' This function calculates conventional AMO index and AMO pattern.
        It is calculated as area-averaged SST anomalies and time averaged 
        SST anomalies in North Atlantic above 0N after removing the Annualcycle 
        climatology and global mean SST (warming trend).

        data: Processed gridded SST data in the shape of (time, lat, lon)
        if not in the order, can be reordered as xarray.transpose
        path (directory): path to data
        var (str): variable name
        start_time (int): start year for the data to be visulaized
        end_time (int): end year for the data to be visulaized
        clim_start: climatology start year
        clim_end: climatology end year (if both None, whole period will be chosen to compute climatology)
        desired: desired variables, can be choosen from 
        ['amo_pattern', 'amo_index'] 
        lat_s = start latitude
        lat_e = end latitude
        lon_s = start longitude
        lon_e = end longitude'''

    if desired==None:
        desired = ['amo_pattern', 'amo_index']

    if to_range==None:
        to_range='0_360'    ## otherwise provide -180_180

    if path is not None and var is not None:

        data = adjust_longitude(rename_dims_to_standard(load_data(path, var, start_time=start_time, end_time=end_time, lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e)), to_range=to_range)
        # data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    elif data is not None:
        data = data

    north_atlantic_sst = data.sel(lat=slice(lat_s, lat_e), lon=slice(lon_s, lon_e))
    

    data_anomalies = calculate_anomaly(north_atlantic_sst, clim_start=clim_start, clim_end=clim_end)
    

    global_mean_data = calculate_global_mean_sst(data, lat_name='lat', lon_name='lon')
    

    amo_index = data_anomalies.weighted(compute_weights(data_anomalies)).mean(('lat', 'lon')) - global_mean_data
    

    # if smoothing_window!=None:
    #     amo_index = amo_index.rolling(time=smoothing_window, center=True).mean().dropna(dim='time')
    

    # amo_index = regional_anomaly
    amo_pattern = xr.cov(calculate_anomaly(data), amo_index, dim='time')/amo_index.var()
    return_desired = []
    if 'amo_pattern' in desired:
        return_desired.append(amo_pattern)
    if 'amo_index' in desired:
        return_desired.append(amo_index)
    
    if len(return_desired)==1:
        return return_desired[0]
    else:
        return return_desired


def compute_nao(path=None, var=None, data=None, clim_start=None, clim_end=None, desired=None, lat_s=None, lat_e=None, rotated=None, use_coslat=None, standardize=None, n_modes=2, to_range=None):

    ''' This function calculates NAO index, NAO pattern and variance fraction.
        It is calculated as Second EOF modes of 500mb Geo-potential height 
        anomaly above 20N after removing the Annualcycle climatology.

        data: Processed gridded SST data in the shape of (time, lat, lon)
        if not in the order, can be reordered as xarray.transpose

        clim_start: climatology start year
        clim_end: climatology end year (if both None, whole period will be chosen to compute climatology)
        desired: desired variables, can be choosen from 
        ['pdo_pattern', 'pdo_index', 'variance_fraction_pdo'] 
        lat_s = start latitude
        lat_e = end latitude
        rotated: if rotated analysis is required. None if unrotated, 
        Varimax and Promax for rotated
        use_coslat: True if None, otherwise give False. It is used to calculated area-averaged EOFs
        standardize = False if None, make it True if you want to standardize
        n_modes: default 2; this will assume that 2nd mode is NAO, change accordingly'''


    if standardize==None:
        standardize=False

    if desired is None:

        desired = ['nao_pattern', 'nao_index', 'variance_fraction_nao']

    if to_range==None:
        to_range='0_360'    ## otherwise provide -180_180

    if path is not None and var is not None:

        data = adjust_longitude(rename_dims_to_standard(load_data(path, var, start_time=start_time, end_time=end_time, lat_s=lat_s, lat_e=lat_e, lon_s=lon_s, lon_e=lon_e)), to_range=to_range)
        # data_anom = calculate_anomaly(data, clim_start=clim_start, clim_end=clim_end)
    elif data is not None:
        data = data

    north_geop = data.sel(lat=slice(lat_s, lat_e))
    

    data_anomalies = calculate_anomaly(north_geop, clim_start=clim_start, clim_end=clim_end)
    

    nao_index = compute_rotated_eofs(data_anomalies, rotated=rotated, n_modes=n_modes, standardize=standardize, use_coslat=use_coslat).scores()[n_modes-1]

    nao_index = nao_index/nao_index.std()
    
    nao_pattern = compute_rotated_eofs(data_anomalies, rotated=rotated, n_modes=n_modes, standardize=standardize, use_coslat=use_coslat).components()[n_modes-1]

    variance_fraction_nao = compute_rotated_eofs(data_anomalies, rotated=rotated, n_modes=n_modes, standardize=standardize, use_coslat=use_coslat).explained_variance_ratio()[1]
    return_desired = []
    if 'nao_pattern' in desired:
        return_desired.append(nao_pattern.squeeze())
    if 'nao_index' in desired:
        return_desired.append(nao_index.squeeze())
    if 'variance_fraction_nao' in desired:
        return_desired.append(variance_fraction_nao.squeeze())
    
    if len(return_desired)==1:
        return return_desired[0]
    else:
        return return_desired


